//
// rhexampRhinoExampleCategory.c++
//
//	Example category.
//
//
//  Copyright (c) 1995, 2000 Silicon Graphics, Inc.  All Rights Reserved.
//  
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of version 2 of the GNU General Public License as
//  published by the Free Software Foundation.
//  
//  This program is distributed in the hope that it would be useful, but
//  WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  
//  Further, this software is distributed without any warranty that it is
//  free of the rightful claim of any third person regarding infringement
//  or the like.  Any license provided herein, whether implied or
//  otherwise, applies only to this software file.  Patent licenses, if
//  any, provided herein do not apply to combinations of this program
//  with other software, or any other product whatsoever.
//  
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write the Free Software Foundation,
//  Inc., 59 Temple Place - Suite 330, Boston MA 02111-1307, USA.
//  
//  Contact information: Silicon Graphics, Inc., 1600 Amphitheatre Pkwy,
//  Mountain View, CA 94043, or http://www.sgi.com/
//  
//  For further information regarding this notice, see:
//  http://oss.sgi.com/projects/GenInfo/NoticeExplan/
//

#ident "$Revision: 1.24 $"

#include <sys/stat.h>
#include <assert.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>

#include <sysadm/format.h>
#include <sysadm/AppContext.h>
#include <sysadm/Log.h>

#include <rhexamp/RhinoExample.h>
#include "rhexampRhinoExampleCategory.h"

namespace rhexamp {

SaCATEGORY_REF_DEF(rhexampRhinoExampleCategory);

//
// Constructor.
//
rhexampRhinoExampleCategory::rhexampRhinoExampleCategory()
: Category("rhexampRhinoExampleCategory")
{
}

//
// Destructor.
//
rhexampRhinoExampleCategory::~rhexampRhinoExampleCategory()
{
    if (_famStarted) {
	FAMClose(&_famConn);
	AppContext::getAppContext().unregisterMonitor(_inputId);
    }
}

//
//  Item rhexampRhinoExampleCategory::createItem(const char *exampleName)
//
//  Description:
//      Respond to a FAM event that indicated that a new example has
//      been created.  This also gets called at startup for each
//      example that already exited.
//      
//      Make sure example is valid, and if so put together Example
//      attributes.
//
//  Parameters:
//      exampleName  Name of the example that showed up.
//
//  Returns:
//	Newly created item corresponding to "exampleName".
//
Item rhexampRhinoExampleCategory::createItem(const char *exampleName)
{
    char exampleFile[PATH_MAX];
    (void)SaStringFormat(exampleFile, sizeof exampleFile,
			 "%s/%s", RHINO_EXAMPLE_DIR, exampleName);

    struct stat f;
    if (stat(exampleFile, &f) == -1) {
	return Item("", "");
    }

    FILE *fp = fopen(exampleFile, "r");
    if (fp == NULL) {
	return Item("", "");
    }

    char buf[100];
    char *type = fgets(buf, sizeof buf, fp);
    (void)fclose(fp);

    if (type == NULL) {
	return Item("", "");
    }

    char *pc = strchr(buf, '\n');
    if (pc) {
	*pc = '\0';
    }

    Item item(getSelector(), exampleName);
    item.setAttr(Attribute(RHINO_EXAMPLE_NAME, exampleName));
    item.setAttr(Attribute(RHINO_EXAMPLE_TYPE, type));
    item.setAttr(Attribute(RHINO_EXAMPLE_MODE, (long long)f.st_mode)); 

    return item;
}

//
//  void rhexampRhinoExampleCategory::processFamEvent(FAMEvent &event)
//
//  Description:
//      Process a single FAM event.
//
//  Parameters:
//      event  The event to process.
//
void rhexampRhinoExampleCategory::processFamEvent(FAMEvent &event)
{
    Log::trace(getSelector(), "Got a fam event");
    switch (event.code) {
    case FAMExists:
    case FAMCreated:
	{
	    Item item(createItem(event.filename));
	    if (item.getSelector() != "") {
		addItem(item);
	    }
	}
	break;
    case FAMChanged:
	{
	    Item item(createItem(event.filename));
	    if (item.getSelector() != "") {
		changeItem(item);
	    }
	}
	break;
    case FAMDeleted:
	removeItem(event.filename);
	break;
    case FAMEndExist:
	endExists();
	break;
    }
}

//
//  void rhexampRhinoExampleCategory::famInput(void* clientData, int, int)
//
//  Description:
//      Input callback that gets called when we get a FAM event.
//
//  Parameters:
//      clientData  ClusterCategory* (this is a static method).
//
void rhexampRhinoExampleCategory::famInput(void* clientData, int, int)
{
    rhexampRhinoExampleCategory* self = (rhexampRhinoExampleCategory*)clientData;

    FAMEvent event;
    while (FAMPending(&self->_famConn) == 1) {
	if (FAMNextEvent(&self->_famConn, &event) != -1) {
	    self->processFamEvent(event);
	}
    }
}

//
//  void rhexampRhinoExampleCategory::startMonitor()
//
//  Description:
//      Set up our FAM connection.
//
//  Returns:
//	0 if successful, -1 if error.
//
void rhexampRhinoExampleCategory::startMonitor()
{
    if (FAMOpen(&_famConn) == 0) {
	_famStarted = true;
	FAMMonitorDirectory(&_famConn, RHINO_EXAMPLE_DIR,
			    &_configDir, NULL);
	_inputId = AppContext::getAppContext().registerMonitor(
	    FAMCONNECTION_GETFD(&_famConn), famInput, this);
    } else {
	endExists();
    }
}

} // namespace rhexamp
