//
// AbstractRhinoExampleTask.java
//
//	Base class for AddRhinoExampleTask and ModifyRhinoExampleTask.
//
//
//  Copyright (c) 1998, 2000 Silicon Graphics, Inc.  All Rights Reserved.
//  
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of version 2 of the GNU General Public License as
//  published by the Free Software Foundation.
//  
//  This program is distributed in the hope that it would be useful, but
//  WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  
//  Further, this software is distributed without any warranty that it is
//  free of the rightful claim of any third person regarding infringement
//  or the like.  Any license provided herein, whether implied or
//  otherwise, applies only to this software file.  Patent licenses, if
//  any, provided herein do not apply to combinations of this program
//  with other software, or any other product whatsoever.
//  
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write the Free Software Foundation,
//  Inc., 59 Temple Place - Suite 330, Boston MA 02111-1307, USA.
//  
//  Contact information: Silicon Graphics, Inc., 1600 Amphitheatre Pkwy,
//  Mountain View, CA 94043, or http://www.sgi.com/
//  
//  For further information regarding this notice, see:
//  http://oss.sgi.com/projects/GenInfo/NoticeExplan/
//

package com.sgi.rhexamp.task;

import java.util.*;
import java.io.*;
import java.text.*;
import com.sgi.sysadm.ui.*;
import com.sgi.sysadm.util.*;
import com.sgi.sysadm.category.*;
import com.sgi.sysadm.ui.taskData.*;
import com.sgi.rhexamp.*;

/**
 * AbstractRhinoExampleTask is the base class for AddRhinoExampleTask
 * and ModifyRhinoExampleTask.
 */
abstract class AbstractRhinoExampleTask extends Task
                                      implements RhinoExampleConstants {

    // These members have package protection so they are accessible by
    // subclasses.
    TaskContext _taskContext;
    TaskData _taskData;
    ResourceStack _rs;
    Category _rhinoExampleCategory;

    static String CLASS_NAME = ResourceStack.getClassName(
	AbstractRhinoExampleTask.class);
    private String _privilege;

    static final String WRITABLE = "writable";

    public AbstractRhinoExampleTask(TaskContext taskContext) {
	super(taskContext);

	_taskContext = taskContext;
	_rs = _taskContext.getResourceStack();
	_taskData = _taskContext.getTaskData();
	_rhinoExampleCategory = taskContext.getHostContext().getCategory(
		RHINO_EXAMPLE_CATEGORY);

	_taskData.setString(RHINO_EXAMPLE_NAME, "");
	_taskContext.appendTaskDataVerifier(RHINO_EXAMPLE_NAME,
					    new TaskDataVerifier() {
	    public void dataOK(int browseFlag, Object context,
			       final ResultListener listener) {
		final String name = _taskData.getString(
		    RHINO_EXAMPLE_NAME).trim();
		if (name.length() == 0) {
		    if (browseFlag == TaskDataVerifier.MUST_BE_SET) {
			ResultEvent event = new ResultEvent(Task.this);
			event.setReason(_rs.getString(
			    CLASS_NAME + ".Error.missingRhinoExampleName"));
			listener.failed(event);
		    } else {
			listener.succeeded(new ResultEvent(this));
		    }
		    return;
		}
		if (name.indexOf('/') != -1) {
		    ResultEvent event = new ResultEvent(Task.this);
		    event.setReason(
			_rs.getString(CLASS_NAME +
				      ".Error.illegalCharInRhinoExampleName"));
		    listener.failed(event);
		    return;
		}
		listener.succeeded(new ResultEvent(Task.this));
	    }
	});

	_taskData.setString(RHINO_EXAMPLE_TYPE, "Clock");
	_taskData.setBoolean(WRITABLE, false);
    }

    /**
     * setOperands verifies that the operands passed to this task are
     * acceptable.
     * 
     * @param operands The operands that this task will operate on.
     * 
     * @exception TaskInitFailedException if operands are not valid.
     */
    public void setOperands(Vector operands) throws TaskInitFailedException {
	if (operands == null || operands.size() == 0) {
	    return;
	}

	if (operands.size() > 1) {
	    throw new TaskInitFailedException(_rs.getString(
		CLASS_NAME + ".Error.tooManyOperands"),
		TaskInitFailedException.INVALID_OPERANDS);
	}

	try {
	    _taskData.setString(RHINO_EXAMPLE_NAME,
				(String)operands.elementAt(0));
	} catch (ClassCastException ex) {
	    throw new TaskInitFailedException(_rs.getString(
		CLASS_NAME + ".Error.invalidOperand"),
		TaskInitFailedException.INVALID_OPERANDS);
	}
    }

    /**
     * ok is called when the user presses the OK button.  runPriv
     * executes a privileged command on the server.
     */
    public void ok() {
	Log.assert(_privilege != null,
		   "You must call setPrivilege() before calling ok().");
	TaskData args = new TaskData();
	args.setAttr(_taskData.getAttr(RHINO_EXAMPLE_NAME));
	args.setAttr(_taskData.getAttr(RHINO_EXAMPLE_TYPE));
	args.setString(RHINO_EXAMPLE_MODE, _taskData.getBoolean(WRITABLE)
		       ? "0644" : "0444");
	try {
	    runPriv(_privilege, args).close();
	} catch (IOException ex) {
	    Log.debug(CLASS_NAME, ex.getMessage());
	    Log.debug(CLASS_NAME, SysUtil.stackTraceToString(ex));
	}
    }

    /**
     * createResultViewPanel is called if the runPriv operation that
     * was started by ok succeeds.  We create a result view showing
     * the newly added item.
     */
    public ResultViewPanel createResultViewPanel() {
	return new ResultViewPanel(
	    _taskContext,
	    _rs,
	    CATEGORY_PACKAGE + RHINO_EXAMPLE_CATEGORY,
	    new ItemTester() {
		public ItemTestResult testItem(Item item) {
		    return item.getSelector().equals(
			_taskData.getString(RHINO_EXAMPLE_NAME)) ?
			ItemTestResult.TEST_PASSED : ItemTestResult.TEST_FAILED;
		}
	    },
	    MessageFormat.format(
		_rs.getString(CLASS_NAME + ".succeeded"),
		new Object[] { _taskData.getString(RHINO_EXAMPLE_NAME) }));
    }

    /**
     * Called by AddRhinoExampleTask and ModifyRhinoExampleTask to set
     * the privilege to be used in the ok() method.
     * 
     * @param privilege Privilege to use in the ok() method.
     */
    void setPrivilege(String privilege) {
	_privilege = privilege;
    }
}
